/****************************************************************************
 *                                                                          *
 * This program called cvi (C verbal interpreter) looks for information     *
 * about variables and data types in an ANSI C program.                     *
 * Copyright (c) 1999 Chepelianskii Alexeii.                                *
 *                                                                          *
 * This program is free software; you can redistribute it and/or modify     *
 * it under the terms of the GNU General Public License as Published by     *
 * the Free Software Fondation; either version 2 of the License, or (at     *
 * your option) any later version.                                          *
 *                                                                          *
 * This program is distributed in the hope it will be useful, but WITHOUT   *
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or    *
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License     *
 * for more details.                                                        *
 *                                                                          *
 * You should have received a copy of the GNU General Public License along  *
 * with this program; if not, write to the Free Software Foundation, Inc.,  *
 * 675 Mass Ave, Cambridge, MA 02139, USA.                                  *
 *                                                                          *
 ****************************************************************************/

#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <stdlib.h>
#include "trad.h"

void static dcl(int, int *cor, int *type,Variable **var, struct Sortie *out);
void static dclabs(int , int * , int * ,Variable **, struct Sortie *);
                                                   /* the same thing */

list *MainL = NULL;  /* To transmit information about the program. 
			A stack is used as trad can be called recursively. */
char *lex = NULL;    /* string with the last token. */
int typelex;         /* type of the last token. */

/* To add a new variable put its name in sorted order
   in that list with its properties : acces modifier ...
   Then add a constant in the CTypes enumeration(trad.h). */
#define infoTypes  { \
       USER      ,  INDULGENT ,  0,   \
       "auto"    ,  STOCK     ,  0,   \
       "char"    ,  TYPE      ,  0,   \
       "const"   ,  ACCSMOD   ,  0,   \
       "double"  ,  TYPE      ,  0,   \
       "enum"    ,  ENUM      ,  0,   \
       "extern"  ,  STOCK     ,  0,   \
       "float"   ,  TYPE      ,  0,   \
       "int"     ,  TYPE      ,  0,   \
       "long"    ,  QUAL      ,  0,   \
       "register",  STOCK     ,  0,   \
       "short"   ,  QUAL      ,  0,   \
       "signed"  ,  QUAL      ,  0,   \
       "static"  ,  STOCK     ,  0,   \
       "struct"  ,  STRUCT    ,  0,   \
       "union"   ,  STRUCT    ,  0,   \
       "unsigned",  QUAL      ,  0,   \
       "void"    ,  TYPE      ,  0,   \
       "volatile",  ACCSMOD   ,  0,    }


/* can define a variable */
const struct Type donnee[] = infoTypes;
/* used for another purpose */
const char *autre[] = { "typedef" };

const char *motcle[] =
{ "auto", "break", "case", "char", "const", "continue" ,"default" ,
  "do", "double", "else", "enum", "extern", "float", "for", "goto",
  "if", "int", "long", "register", "return", "short", "signed",
  "sizeof", "static", "struct", "switch", "typedef", "union",
  "unsigned", "void", "volatille", "while" };

int const NCLES = sizeof(donnee) / sizeof(donnee[0]); 
int const NREST = sizeof(motcle) / sizeof (motcle[0]);
int const NAUTRE = sizeof(autre) / sizeof (autre[0]);

int IsType(char *s)
{
      return dichoType(s, donnee, NCLES) != NULL || 
	dicho(s, autre, NAUTRE) || consulter(s, TP_INF) != NULL;
}


/* some inline function to simpliy procedures
   whch would be too long otherwise */
inline static void initSortie(Sortie *p, int q)
{
    p->var.nom = NULL;
    p->var.dec = NULL;
    p->var.text = NULL;
    p->erreurpos = 0;
    if (q) {
       p->var.type = NULL;
       p->var.strtype = NULL;
       p->var.str = NULL;
    }
    p->quelvar[FIRST] = UNDEF;
    p->quelvar[MIDDLE] = UNDEF;
    p->quelvar[LAST] = UNDEF;
}


void copierVar(Variable **var, Variable *p)
{
    if (p != NULL) {
       copierVar(var, p->gauche);
       *var = ajoutarbre(*var, &p->var);
       copierVar(var, p->droit);
    }
}


/* to treat errors when using the void type */
static inline void VoidError(const Sortie *sr, int *cor)
{
    if (sr->erreurpos && (sr->quelvar[MIDDLE] == VAR || 
			     sr->quelvar[MIDDLE] == TABLEAU)) {
       error("%s %s has the void type", 
	     (sr->quelvar[MIDDLE] == VAR) ? "variable" : "array",
	     sr->var.nom);
       *cor = 0;
    }
}


inline static int catchType(struct Sortie *dec, int *tr, int *cor, char *func,
		                         Variable **fd, struct Type *typevar)
{
    do {
       /* to check if a typedef is correct :
	  use it for an arbitrary variable */
       dcl(DECL, cor, tr, fd, dec);
       VoidError(dec, cor);
       if (!*cor || *tr != DECLARATION || (typelex != ';' && typelex != ',')) {
	  error(" incorect typedef line %d", nlignes);
	  return 0;
       }
       if (defineType(dec, func, typevar, NCLES, nblocs) == NULL)
	  return ERREUR;
    } while (typelex == ',');
    return TYPEDEF;
}

void remettre(int c)
{
    if (c == NOM)
       remettrech(lex);
    else
       remettrecar(c);
}


/* translate a C declaration/definition in a readable string */
int trad(Variable **var, struct infovar *nomvar, char *func)
{
    int correct = OUI;
               /* indique si la chaine traitee represente un 'cast' */
    int traitertype = OUI;
    int trouv = DECLARATION;

    struct Type typevar[] = infoTypes;
    struct Sortie decVar;
    Variable *found = NULL;
    Prog *prg = (Prog *) pushG(sizeof(Prog), &MainL);
    
    prg->var = var;
    prg->func = func;
    
    do {
       if (traitertype)
	  lirelex();
       initSortie(&decVar, traitertype);
       if (traitertype) {
          correct = liretype(DECL, typevar, NCLES, &decVar, func);
	  if (correct == DEFSTRUC) {
	     nomvar->nom = nomvar->type = decVar.var.type;
	     return DEFSTRUC;
	  } else if (correct == TYPEDEF)
	     return catchType(&decVar, &trouv, &correct, func, &found,typevar);
       }
       dcl(DECL, &correct, &trouv, &found, &decVar);
       if (trouv == CAST) {
	  if (typelex != ')')
	     trouv = ERREUR;
	  return trouv;
       }
       if (typelex == ',')
	  traitertype = NON;
       else if (typelex == ';') {
	  traitertype = OUI;
       } else if (typelex == '{' && decVar.quelvar[FIRST] == FONCTION) {
	  if (!decVar.arg) {
	     error(" parameter name omitted line %d", nlignes);
	     return 0;
	  }
	  nblocs++;
	  copierVar(var, found);
	  trouv = MENTION;
       } else if (typelex == ':')  { /* binary fields */
	 #define MAX_SIZE 50
	 char buf[MAX_SIZE];
	 int c, i;
	 /* 
	    It would be really useless to use dynamic memory for this.
	    Generally dynamical memory is used for output strings 
	    because it is difficult to set their dimension in the program.
	 */
	 if (decVar.quelvar[MIDDLE] != VAR) {
	    error(" bit field founded on a derived type");
	    return 0;
	 }
         mstrcat(&decVar.var.strtype, ": bit field of ");
	 mstrcat(&decVar.var.type, "bit field ");
	 for (i = 0 ; (c = lirecar()) != ';' && c != EOF && i + 1 < MAX_SIZE 
		 ; i++)                                /* for '\0' */
	    buf[i] = c;
	 if (i >= MAX_SIZE - 1) {
	    fprintf(stderr, " warning : too long size requested for %s "
		    "line %d\n", decVar.var.nom, nlignes);
	    return 0;
	 }
	 buf[i] = '\0';
	 if (c != ';' || !(i = atoi(buf)))   /* new use of i. */
	    return 0;      /* Will be treated as a syntax error. */
	 vastrcat(3, &decVar.var.strtype, buf, " bite", i == 1 ? " " : "s ");
	 mstrcat(&decVar.var.type, buf);
	 #undef MAX_SIZE
       } else if (trouv == OLD_MENTION) 
	 remettre(typelex);
       else {
	  correct = 0;
          error(" syntax");
       }
       VoidError(&decVar, &correct);
       if (correct) {
	  *nomvar = decVar.var;

	  if (decVar.quelvar[FIRST] == FONCTION)
	     nomvar->str = NULL;

	  nomvar->fonction = (decVar.quelvar[FIRST] == FONCTION ||
				typevar[Extern].ntrouve) ? EXTERN :  func;
	  nomvar->ligne = nlignes;
	  nomvar->status = nblocs;
	  if (trouv == MENTION)
	     nomvar->status--;
	  nomvar->whatIs = decVar.quelvar[1];
	  if (decVar.quelvar[FIRST] == FONCTION)
	     nomvar->def  = (trouv == MENTION) ? 1 : 0;
	  else if (typevar[Extern].ntrouve)
	     nomvar->def = 0;
	  else
	     nomvar->def = 1;
	  *var = ajoutarbre(*var, nomvar);
       } else
          return ERREUR;
    } while (!traitertype);
    popG(&MainL);
    return trouv;
}

/* check for a variable-length parameter list */
inline static void VariableParameterList(int *cor, int *narg, int *type, 
				  int *VoidSeul, struct Sortie *res)
{
    int variab = 0;
    int npoint = 0;

    while ((*type = lirelex()) == '.')
       if (++npoint == 3) {
	  if (!*VoidSeul) {
	     error(" void in variable-length"
		   " parameter list line %d", nlignes);
	     break;
	  }
	  mstrcat(&res->var.text, " variable-length parameter list ");
	  mstrcat(&res->var.dec, "var l");
	  variab = 1;
	  narg++;
	  break;
       }
    if (npoint && !variab) {
       fprintf(stderr, " unexpected '.' line %d\n", nlignes);
       *cor = 0;
       return;
    }
    if (variab) {
       *type = lirelex();
       return;
    }
}

inline void decal(struct Sortie *res, int val)
{
    if (res->quelvar[1] != val)
       res->quelvar[0] = res->quelvar[1];
    res->quelvar[1] = val;
    if (!res->quelvar[2] || res->quelvar[2] == VAR)
       res->quelvar[2] = val;
}


/* dcl : analyse un declarateur absolu */
/* dic indice dans les tableaux de sortie */
void static dcl(int dic, int *cor, int *trouv, Variable **racine,
		struct Sortie *res)
{
    int ne;
    for (ne = 0; lirelex() == '*' ; )
       ne++;
    dclabs(dic, cor, trouv, racine, res);
    while (ne-- > 0) {
       mstrcat(&res->var.dec, " pt sr ");
       mstrcat(&res->var.text, " pointer to ");
       decal(res, POINTEUR);
    }
}

/* dclabs : analyse un declarateur absolu, trouv 
   indique si la variable lue est un' cast' */
void static dclabs(int dic, int *cor, int *trouv, Variable **racine,
		   struct Sortie *res)
{
    int type;
    int dejalu = 0;
    
    if (typelex == '(') {
       dcl(dic, cor, trouv, racine, res);
       if (typelex != ')')
	  error(") missing");
    } else if (typelex == NOM) {
       if (dicho(lex, motcle, NREST))
	  error("%s is a keyword in C", lex), *cor = 0;
       mstrcpy(&res->var.nom, lex);
       if (dic == DECL + 1)
          res->arg++;
       decal(res, VAR);
    } else if (dic == DECL) {
       *trouv = CAST;
       decal(res, VAR);
    } else if (dic > DECL) {
       if (res->erreurpos && (typelex == ')' || typelex == ',' ) && 
	   res->quelvar[1] == POINTEUR ) {
          res->erreurpos = 0;
	  remettrecar(typelex);
       } else if (typelex == CROCHETS)
	  remettrech(lex);
       else
	  remettrecar(typelex);
       decal(res, VAR);
    } else
       error("a variable name or a (dcl) expected"), *cor = 0;
    if (typelex == ')' && *trouv == CAST)
       return;
    while ((type = lirelex()) == PARENS || type == CROCHETS 
	   || type == INDEFPARENS || type == OLDPARENS)
       if (type == PARENS || type == INDEFPARENS || type == OLDPARENS) {

	  int narg = 0;
	  int VoidSeul = 1;
	  struct Sortie newone;

	  initSortie(&newone, OUI);
	  if (!dejalu)
	     dejalu = PARENS;
          else
	     error("incorrect function definition"), *cor = 0;
	  mstrcat(&res->var.dec, " fnc");
	  vastrcat(2, &res->var.text, " function f:", conv(dic));
	  if (type != INDEFPARENS) {
	     mstrcat(&res->var.dec, " of");
	     mstrcat(&res->var.text, " of (");
	  }
	  if (type == OLDPARENS && dic == 1) {
	     char *result = NULL;
	     int ressize = 0;

	     mstrcat( &res->var.text, " ");
	     while ((typelex = liremot(&result, &ressize)) != ')')
	        vastrcat(2, &res->var.text, result, " ");
	     free((void *) result);
	     *trouv = OLD_MENTION;
	  } else if (type == OLDPARENS) {
	     *cor = NON;
	  } else {
	     do {
	        struct Type typevar[] = infoTypes;
		
		if (narg) {
		   initSortie(&newone, OUI);
		   VariableParameterList(cor, &narg, &type, 
					 &VoidSeul, res);
		   if (type == ')')
		      break;
		} else
		   lirelex();
		if (type == INDEFPARENS) {
		   res->arg++;
		   break;
		}
		*cor = liretype(dic+1, typevar, NCLES, &newone , res->var.nom);
		dcl(dic+1, cor, trouv, racine, &newone);
	        vastrcat(5, &res->var.text, " ", newone.var.nom, 
			 newone.var.nom != NULL ? ":" : " ", 
			 newone.var.text, " ");
		mstrcat(&res->var.dec, newone.var.dec);
		if (newone.arg)
		   res->arg = newone.arg;
		if (narg > 0 && !VoidSeul) {
		   error("`void' in parameter list must be the entire list\n");
		   *cor = 0;
		   return;
		}
		if (typevar[Extern].ntrouve || typevar[Static].ntrouve
		    || typevar[Auto].ntrouve) {
		   error("function argument with a storage-class"
			 " line %d\n",nlignes);
		   *cor = 0;
		}
		vastrcat(2, &res->var.text, newone.var.strtype, 
			 (typelex == ',') ? "," : "");
		mstrcat(&res->var.dec, newone.var.type);
		/* if the function has no arguments */
		if ( !newone.var.dec && !newone.var.nom && newone.erreurpos) {
		   if (!narg) {
		      newone.erreurpos = 0;
		      VoidSeul = 0;
		   }  
		}
		VoidError(&newone, cor);
		if (newone.var.nom && dic <= DECL + 1) {
		   struct infovar varlue;
		   varlue = newone.var;
		   varlue.fonction = res->var.nom;
		   varlue.ligne = nlignes;
		   varlue.def = (res->quelvar[1] != FONCTION) ? 1 : 0;
		   varlue.status = 1;
		   *racine = ajoutarbre(*racine, &varlue);
		}
		narg++;
	     } while (typelex == ',');
	  }
	  if (narg == 1 && !VoidSeul)
	     res->arg++;
	  if (typelex != ')')
	     error(") missing"), *cor = 0;
	  vastrcat(3, &res->var.text, ") f:", conv(dic), " returning ");
          mstrcat(&res->var.dec, " retr");
	  if (*trouv != OLD_MENTION)
	     decal(res, FONCTION);
       } else {
	  if (!dejalu)
	     dejalu = CROCHETS;
          else if (dejalu == PARENS)
	     error("incorrect array definition"), *cor = 0;
	  vastrcat(3, &res->var.text, " array", lex, " of ");
	  mstrcat(&res->var.dec, " pt ");
	  decal(res, TABLEAU);
       }
}

#undef infoTypes

static int dcomp(char **s1, char **s2) { return strcmp(*s1, *s2); }
int dicho(char *mot, const char *tvar[], int n)
{
    if (bsearch((void *)&mot, (void *)tvar,n, sizeof(tvar[0]), (ARG)dcomp)
	   == NULL)
       return 0;
    return 1;
}


int anaind(Variable **var, char *func, char **mot, int *size)
{
    char *temp = 0;
    int c;
    int sizetmp = 0;

    mistrcat(mot, "[", size);
    while ((c = liremot(&temp, &sizetmp)) != ']' && c != EOF) {
       mistrcat(mot, temp, size);
       unknownvar(var, func, &temp, size);
    }
    mistrcat(mot, "] ", size);
    free((void *) temp);
    return c;
}


/* interface for decl.c */
static int anainit(void)
{
       int c;
       int numblocs = 0;
       int nparens = 0;
       static int size = 0;
       static char *mot = NULL;
       Prog *pr = (Prog *) getG(MainL);

       while ((c = liremot(&mot, &size)) != ';' && c != EOF && 
	       (numblocs || c != ',' || nparens )) {
	  if (c == '{')
	     numblocs++;
          else if (c == '}')
	     numblocs--;
          else if (c == '(')
	     nparens++;
          else if (c == ')')
	     nparens--;
	  else
	     unknownvar(pr->var, pr->func, &mot, &size);
	  if (nparens < 0 || numblocs < 0)
	     break;
       }
       return c;
}


/* read a word from the imput returns its type */
static int ReadToken(void)
{
    int c, fonc = 0;
    static maxlex = 300;          /* size of lex */
    static const step = 100;
    char *p;

    if (lex == NULL)
       lex = (char *) malloc (sizeof(char) * maxlex);
    p = lex;
    while (negligerComment(c = lirewspace()))
       ;
    if (c == '(') {
       int i;
       c = lirewspace();
       remettrecar(c);
       if (c == ')')
	  return INDEFPARENS;
       for (i = 0 ; isalpha(c = lirecar()) || c == '_' || c == SPECIAL ; i++) {
	   while (p - lex + 1 >= maxlex)
	      lex = (char *) realloc((void *) lex, maxlex += step);
	   lex[i] = c;
       }
       lex[i] = '\0';
       remettrecar(c);
       if (i) 
	  remettrech(lex);
       if (IsType(lex)) {
	  fonc = 1;
	  return PARENS;
       }
       if (!i)
	  return typelex = '(';
       return OLDPARENS;
    } else if (c == '[') {
       Prog *pr = (Prog *) getG(MainL);

       *lex = '\0';
       if (anaind(pr->var, pr->func, &lex, &maxlex) != ']')
	  return EOF;
       return typelex = CROCHETS;
    } else if (c == '=') {
       return typelex = anainit();
    } else if (isalpha(c) || c == '_' || c == SPECIAL) {
       /* 
	  The SPECIAL letter can begin a variable name to avoid the confusion
	  between the variables definied by the program and those by the user.
       */
       for (*p++ = c; isalnum(c = lirecar()) || c == '_'; )
	   *p++ = c;
       *p = '\0';
       remettrecar(c);
       return typelex = NOM;
    } else
       return typelex = c;
}    

int lirelex(void)
{
    int type;

    for ( ; ; ) {
       type = ReadToken();
       if (type == NOM)
	  if (OpenMacro(lex))
	     continue;
       break;
    }
    return type;
}

/* useful functions */
static char chif[30];

/* convert n into a string : sprintf can be used instead. */
char *conv(int n)
{
     static char *p = chif;
     static j = 0;
     
     if (n < 0)
        *p++ = '-', n = -n;
     if (n / 10)
        j++, conv(n / 10);
     *p++ = (n % 10 +'0');
     if (j == 0) {
        *p = '\0';
	p = chif;
	return chif;
     } else
	j--;
     return NULL;        /* avoid bogus warnings */
}

/* dupchaine : duplicates a string in the RAM memory */
char *dupchaine(char *s)
{
    char *p;

    p = (char *) malloc(strlen(s) + 1);
    if (p != NULL)
       strcpy(p, s);
    else
       AllocError();
    return p;
}

static char *mistr(char **s1, const char *s2, int *n, 
		                          char *(*f)(char *, const char *))
{
     const step = 100;
     int sz2 = (s2 != NULL) ? strlen(s2) : 0;
     int sz1 = (*s1 != NULL) ? strlen(*s1) : 0;

     if (sz2 + sz1 + 1 < *n && ( *s1 = (char *) realloc((void *) *s1, 
				   *n = sz1 + sz2 + step +1)) == NULL)
        AllocError();
     return (*f)(*s1, s2);
}

char *mistrcat(char **s1, const char *s2, int *n)
{
     return mistr(s1, s2, n , strcat);
}

char *mistrcpy(char **s1, const char *s2, int *n)
{
     return mistr(s1, s2, n , strcpy);
}

static char *resize(char **s, const char *v, char *(*f)(char *, const char *) )
{
    if (*s == NULL && v != NULL) {
       if ((*s = (char *) malloc (sizeof(char) * (strlen(v) + 1))) == NULL)
                                                  /* for '\0' */
	  AllocError();
       **s = '\0';
    } else if (v != NULL && *s != NULL)
       if ((*s = (char *) realloc ((void *) *s, 
		      sizeof(char)*(strlen(v) + strlen(*s) +1))) == NULL)
	  AllocError();
    if (v != NULL && *s != NULL)
       (*f)(*s, v);
    return *s;
} 

char *mstrcat(char **s, const char *v)
{
    return resize(s, v, strcat);
}


char *mstrcpy(char **s, const char *v)
{
    return resize(s, v, strcpy);
}

#include <stdarg.h>

/* to avoid the waste of time for memory allocation
   when I format a simple string : 
   warning : vastrcat only treat the char * type.
*/
char *vastrcat(int narg, char **f, ...)
{
     va_list pa;
     char *p;
     int size;
     int i;

     size = (*f == NULL) ? 0 : strlen(*f);

     va_start(pa, f);
     for (i = 0; i < narg; i++) {
       if ((p = va_arg(pa, char * )) != NULL)
	   size += strlen(p);
     }
     size++; /* for '\0' */
     
     va_start(pa, f);
     if (*f != NULL)
        *f = (char *) realloc ((void *) *f, size * sizeof(char));
     else if (size) {
        *f = (char *) malloc (size * sizeof(char));
	**f = '\0';
     }
     for (i = 0; i < narg ; i++) {
       if ((p = va_arg(pa, char * )) != NULL)
	   strcat(*f, p);
     }
     va_end(pa);
     return *f;
}

/* maketable : makes columns, the numeration begins from the
               left margin, strings are supposed to contain
               no \t. */
void maketable(int ncolumns, int leftmargin, int rightmargin, ...)
{
    va_list pa;
    int margin, outstr, i, stop;
    char **strlist = (char **) malloc(ncolumns * sizeof(char *));
    int *width  = (int *) malloc(ncolumns * sizeof(int));
    int *curpos = (int *) malloc(ncolumns * sizeof(int));
    int *length = (int *) malloc(ncolumns * sizeof(int));
    int *enough = (int *) malloc(ncolumns * sizeof(int));

    if (strlist == NULL || width == NULL || curpos == NULL || length == NULL)
       AllocError();
    va_start(pa, rightmargin);
    /*
        the int value is the width of the column and char *
	points to the string that will be put in.
     */
    /* finding these values ... */
    for (i = 0; i < ncolumns ; i++) {
       width[i] = va_arg(pa, int);  
       strlist[i]  = va_arg(pa, char *);
    }
    for (i = 0; i < ncolumns ; i++) {
       length[i] = strlen(strlist[i]);
       curpos[i] = 0;
    }

    /* printing the columns ... */
    while (1) {
       for (i = 0; i < leftmargin; i++)
	  putchar(' ');
       for (i = 0; i < ncolumns ; i++) {
	  register j;
	  /* the string is already wrote */
	  if (curpos[i] >= length[i]) {
	     for (j = 0 ; j < width[i] + rightmargin ; j++)
	        putchar(' ');
	     enough[i] = 1;
	  /* there is enough space */
	  } else if (
	      (j = width[i] - rightmargin - length[i] + curpos[i] + 1) > 0) {
	     int count;
	     j += rightmargin;
	     for (count = curpos[i]; count < length[i]; count++)
	        putchar(strlist[i][count]);
	     while (j--)
	        putchar(' ');

	     curpos[i] = length[i];
	     enough[i] = 1;
	  /* if no lets break the line ... */
	  } else {
	     int count;
	     j = width[i] - rightmargin;
	     while (j > 0) {
	        if (isspace(strlist[i][j+curpos[i]]))
		   break;
		j--;
	     }
	     if (!j)
	        j = width[i] - rightmargin;
	     count = curpos[i];
	     curpos[i] += j;
	     for (j = count; j < curpos[i]; j++)
	        putchar(strlist[i][j]);
	     /* skip the space */
	     if (isspace(strlist[i][curpos[i]]))
	        curpos[i]++;
	     for (j = 0; j < width[i] +rightmargin -curpos[i] +count; j++)
	        putchar(' ');
	     enough[i] = 0;
	  }
       }
       putchar('\n');
       for (i = 0; i < ncolumns; i++) {
	  if (!enough[i]) {
	     stop = 0;
	     break;
	  } else
	     stop = 1;
       }
       if (stop)
	  break;
    }

    va_end(pa);
    free((void *) strlist);
    free((void *) width);
    free((void *) curpos);
    free((void *) length);
    free((void *) enough);
}


/* The unconsidered use of that macro can cause problems not easy to
   detect : be careful ! */
#define Error(args, fmt) { va_start(args, fmt); \
                    fprintf(stderr, " error : "); \
                    vfprintf(stderr, fmt, args); \
                    va_end(args); }

static const int MAXERROR = 15;
static int nerrors = 0;

int iserror(void)
{
     return nerrors;
}


void error(char *fmt, ...)
{
     va_list args;
     Error(args, fmt);
     fprintf(stderr, " File %s\n", filename);
     if (++nerrors > MAXERROR)
        fatalerror("confused by errors stop parsing");
}

void fatalerror(char *fmt, ...)
{
     va_list args;
     Error(args, fmt);
     fprintf(stderr, "\n");
     exit(1);
}

#undef Error

void AllocError(void)
{
     error("can not allocate memory");
     exit(1);
}


/* Compare s1 and s2 using words as unit of comparison. */
int motcmp(char *s1, char *s2)
{
     static char *mot = NULL;
     static int size = 0;
     register res = 0;

     remettrecar(0);   /* to know when s1 is finished */
     remettrech(s1);
     
     while (liremot(&mot, &size))
        if (!strcmp(mot, s2))
	   res++;
     return res;
}

